// GpibGsTi9914.cc

/********************COPYRIGHT NOTIFICATION**********************************
This software was developed under a United States Government license
described on the COPYRIGHT_UniversityOfChicago file included as part
of this distribution.
****************************************************************************/

// Author: John Winans
// Date: 12/94
// Change from Hideos to MPF: Marty Kraimer July 1998

// Support for the IP-488 (Green Springs GPIB IP using the ti9914)

#include <vxWorks.h>
#include <iv.h>
#include <stdlib.h>
#include <stddef.h>
#include <string.h>
#include <stdio.h>
 
#include <semLib.h>
#include <tickLib.h>
#include <taskLib.h>
#include <intLib.h>
#include <logLib.h>

#include "WatchDog.h"
#include "Reboot.h"
#include "DLList.h"
#include "IndustryPackModule.h"
#include "Gpib.h"

class GpibGS : public Gpib
{
public:
    GpibGS(const char *moduleName, IndustryPackModule *pIndustryPackModule,
        int intVec);
    static GpibGS *findGpibGS(const char *moduleName);
    virtual int cmd(const char *buf, int cnt, double time_limit);
    virtual int write(const char *buf, int cnt,
	int address, double time_limit);
    virtual gpibStatus read(char *buf, int cnt, int *actual,
	int address, double time_limit, int eos);
private:
    static void interruptHandler(void*);
    void sicGpib(void);
    int cmdGpib(const char *buf, int cnt, double time_limit);
    int writeGpib(const char *buf, int cnt, bool raw_flag, double time_limit);
    int readGpib(char *buf, int cnt, double time_limit, int eos);
    static void rebootCallback(void *);
    
    IndustryPackModule *pIPM;		// my module descriptor
    const char         *name;
    int                intVec;
    volatile struct ip488RegisterMap	*regs;	// hardware registers
    unsigned short     copyIsr0; // can be altered in the IRQ handler
    unsigned short     copyIsr1; // can be altered in the IRQ handler
    int	               transferSize;
    unsigned char      *transferBuffer;
    bool	       transferEoiOnLast;
    enum	{IP488_READ, IP488_WRITE, IP488_IDLE} transferState;
    int   transferEos;

};
 
// following prevents infinite loop waiting for cmdIP488_AUXxxx to complete
static int cmdIP488_AUXretrys = 100;

static char *untalkUnlisten = "_?";

extern "C" {
int gpibGsTi9914SetDelay(int delay){ cmdIP488_AUXretrys = delay;return(0);}
}


int GpibGsTi9914Debug=0;

typedef struct  ip488RegisterMap
{
  unsigned short intReg0;
  unsigned short intReg1;
  unsigned short addressStatus;
  unsigned short cmdStatus;
  unsigned short address;
  unsigned short serialPoll;
  unsigned short commandPasss;
  unsigned short data;
  unsigned short addressSwitch;
  unsigned short vectorRegister;
} ip488RegisterMap;

#define GREEN_SPRING_ID  0xf0
#define GSIP_488  0x14

// Interrupt Status Register 0 Bits
//      Name            Bit(s)       register
#define isr0END         (1<<3)		// EOI present on bus
#define isr0BO          (1<<4)		// Ready to write a byte
#define isr0BI          (1<<5)		// Ready to read byte

// Interrupt Status Register 1 Bits
#define isr1SRQ         (1<<1)		// SRQ on bus is active
#define isr1ERR		(1<<6)		// handshake error (no listener)

// Address Status Register Bits
#define asrATN		(1<<5)		// ATN is true (low) on bus
#define asrLADS		(1<<2)		// TNS9914 is primary listener
#define asrTADS		(1<<1)		// TNS9914 is primary talker

#define IP488_CIC_ADDR  0       // Primary address of the TI9914 controller
 
/* TI9914 Auxiliary Commands */
#define IP488_AUXswrstC    0x00    /* Software reset CLEAR              */
#define IP488_AUXswrstS    0x80    /* Software reset SET                */
#define IP488_AUXrhdf      0x02    /* Release RFD holdoff               */
#define IP488_AUXhdfeS     0x84    /* Holdoff on EOI only SET           */
#define IP488_AUXnbaf      0x05    /* New byte available false          */
#define IP488_AUXfeoi      0x08    /* Send EOI with next byte           */
#define IP488_AUXlonC      0x09    /* Listen only CLEAR                 */
#define IP488_AUXlonS      0x89    /* Listen only SET                   */
#define IP488_AUXtonC      0x0A    /* Talk only CLEAR                   */
#define IP488_AUXtonS      0x8A    /* Talk only SET                     */
#define IP488_AUXgts       0x0B    /* Go to standby                     */
#define IP488_AUXtca       0x0C    /* Take control asynchronously       */
#define IP488_AUXsicC      0x0F    /* Send interface clear CLEAR        */
#define IP488_AUXsicS      0x8F    /* Send interface clear SET          */
#define IP488_AUXsreC      0x10    /* Send remote enable CLEAR          */
#define IP488_AUXsreS      0x90    /* Send remote enable SET            */
#define IP488_AUXdaiC      0x13    /* Disable all interrupts CLEAR      */
#define IP488_AUXdaiS      0x93    /* Disable all interrupts SET        */

void cmdIP488_AUXlonC(volatile ip488RegisterMap *regs)
{
    // Is called by interruptHandler; Must not call printf
    int i;
    unsigned short status = 0;
    for(i=0; i<cmdIP488_AUXretrys; i++) {
        status = regs->addressStatus;
        if(!(status&asrLADS)) break;
        regs->cmdStatus = IP488_AUXlonC;
    }
    if(GpibGsTi9914Debug && (i >1))
        logMsg("cmdIP488_AUXlonC n times %d status 0x%2.2hx\n",
            i,(int)status,0,0,0,0);
}

bool cmdIP488_AUXlonS(volatile ip488RegisterMap *regs)
{
    int i;
    unsigned short status = 0;
    for(i=0; i<cmdIP488_AUXretrys; i++) {
        status = regs->addressStatus;
        if(status&asrLADS) break;
        regs->cmdStatus = IP488_AUXlonS;
    }
    if(GpibGsTi9914Debug && (i >1))
        printf("cmdIP488_AUXlonS n times %d status 0x%2.2hx\n",i,(int)status);
    return((i<cmdIP488_AUXretrys) ? true : false);
}

void cmdIP488_AUXtonC(volatile ip488RegisterMap *regs)
{
    // Is called by interruptHandler; Must not call printf
    int i;
    unsigned short status = 0;
    for(i=0; i<cmdIP488_AUXretrys; i++) {
        status = regs->addressStatus;
        if(!(status&asrTADS)) break;
        regs->cmdStatus = IP488_AUXtonC;
    }
    if(GpibGsTi9914Debug && (i >1))
        logMsg("cmdIP488_AUXtonC n times %d status 0x%2.2hx\n",
            i,(int)status,0,0,0,0);
}

bool cmdIP488_AUXtonS(volatile ip488RegisterMap *regs)
{
    int i;
    unsigned short status = 0;
    for(i=0; i<cmdIP488_AUXretrys; i++) {
        status = regs->addressStatus;
        if(status&asrTADS) break;
        regs->cmdStatus = IP488_AUXtonS;
    }
    if(GpibGsTi9914Debug && (i >1))
        printf("cmdIP488_AUXtonS n times %d status 0x%2.2hx\n",i,(int)status);
    return((i<cmdIP488_AUXretrys) ? true : false);
}

bool cmdIP488_AUXgts(volatile ip488RegisterMap *regs)
{
    int i;
    unsigned short status = 0;
    for(i=0; i<cmdIP488_AUXretrys; i++) {
        status = regs->addressStatus;
        if(!(status&asrATN)) break;
        regs->cmdStatus = IP488_AUXgts;
    }
    if(GpibGsTi9914Debug && (i >1))
        printf("cmdIP488_AUXgts n times %d status 0x%2.2hx\n",i,(int)status);
    return((i<cmdIP488_AUXretrys) ? true : false);
}

bool cmdIP488_AUXtca(volatile ip488RegisterMap *regs)
{
    int i;
    unsigned short status = 0;
    for(i=0; i<cmdIP488_AUXretrys; i++) {
        status = regs->addressStatus;
        if(status&asrATN) break;
        regs->cmdStatus = IP488_AUXtca;
    }
    if(GpibGsTi9914Debug && (i >1))
        printf("cmdIP488_AUXtca n times %d status 0x%2.2hx\n",i,(int)status);
    return((i<cmdIP488_AUXretrys) ? true : false);
}

int initGpibGsTi9914(const char *moduleName, const char *carrierName,
    const char *siteName, int intVec)
{
    if(Gpib::find(moduleName)) {
	printf("initGpibGsTi9914 %s already exists\n",moduleName);
	return(0);
    }
    IndustryPackModule *pIPM = IndustryPackModule::createIndustryPackModule(
	moduleName,carrierName,siteName);
    if(!pIPM) return(0);
    unsigned char manufacturer = pIPM->getManufacturer();
    unsigned char model = pIPM->getModel();
    if(manufacturer!=GREEN_SPRING_ID) {
	printf("initGpibGsTi9914 manufacturer 0x%x not GREEN_SPRING_ID\n",
	    manufacturer);
	return(0);
    }
    if(model!=GSIP_488) {
       printf("initGpibGsTi9914 model 0x%x not a IP-488\n",model);
       return(0);
    }
    new GpibGS(moduleName,pIPM,intVec);
    return(0);
}

GpibGS::GpibGS(const char *name, IndustryPackModule *pIndustryPackModule,
      int intVecIn)
: pIPM(pIndustryPackModule), intVec(intVecIn), 
  transferSize(0), transferBuffer(0),
  transferEoiOnLast(false), transferState(IP488_IDLE), transferEos(0)
{
    registerName(name);
    if((intVec&1)==0) {
        printf("%s intVec MUST be odd. It is changed from %d to %d\n",
            name, intVec, intVec|1);
        intVec |= 1;
    }
    // Find the register map of the IP card
    regs=(ip488RegisterMap *)pIPM->getMemBaseIO();
    if(GpibGsTi9914Debug)printf("GpibGS regs %p\n",regs);
    // configure the interrupts
    pIPM->intConfig(0);
    if(intConnect(INUM_TO_IVEC(intVec),(VOIDFUNCPTR)interruptHandler,(int)this)){
	printf("GpibGsTi9914 %s intConnect failed\n",getGpibName());
    }
    regs->vectorRegister = (unsigned short)intVec;
    // Issue a software reset to the GPIB controller chip
    regs->cmdStatus = IP488_AUXswrstS;
    // Set the ti9914's address to use while it is CIC
    regs->address = IP488_CIC_ADDR;
    // Toss interrupt status values before we release the reset
    copyIsr0 = regs->intReg0;
    copyIsr1 = regs->intReg1;
    regs->cmdStatus = IP488_AUXswrstC;
    regs->intReg0 = isr0BO|isr0BI;
    regs->intReg1 = isr1ERR;
    sicGpib();				// assert IFC
    regs->cmdStatus = IP488_AUXsreS;
    regs->cmdStatus = IP488_AUXdaiC;
    // Prime our copies of these registers
    copyIsr0 = regs->intReg0;
    copyIsr1 = regs->intReg1;
    copyIsr0 = copyIsr0 | isr0BO;
    pIPM->intEnable(0);
    Reboot::rebootHookAdd(rebootCallback,(void *)this);
}

int GpibGS::cmd(const char *buf, int cnt, double time_limit)
{
  if(GpibGsTi9914Debug) {
    printf("GpibGS::cmd cnt %d time_limit %f[",cnt,time_limit);
    for (int i=0; i<cnt; i++) printf("%3.2x",buf[i]);
    printf("]\n");
  }
  int actual = cmdGpib(buf,cnt,time_limit);
  if(GpibGsTi9914Debug) printf("GpibGS::cmd actual(%d)\n",actual);
  return(actual);
}

int GpibGS::write(const char *buf, int cnt, int addr, double time_limit)
{
    if(GpibGsTi9914Debug) {
        printf("GpibGS::write addr %d cnt %d time_limit %f[",
            addr,cnt,time_limit);
        for (int i=0; i<cnt; i++) printf("%3.2x",buf[i]);
        printf("]\n");
    }
    char cmdbuf[3];
    strcpy(cmdbuf,untalkUnlisten);
    cmdbuf[2] = (char)(addr + 0x20);
    int count = cmdGpib(cmdbuf,3,time_limit);
    if(count != 3 && GpibGsTi9914Debug) {
        printf("GpibGS::write cmdGpib error before count %d\n",count);
        return(-1);
    }
    int actual = writeGpib(buf,cnt,0,time_limit);
    if(actual!=cnt || ioStatus!=gpibSuccess) {
        if(GpibGsTi9914Debug)
            printf("GpibGS::write actual %d ioStatus %d\n",actual,ioStatus);
        return(-1);
    }
    count = cmdGpib(untalkUnlisten, 2, time_limit);
    if(count !=2 && GpibGsTi9914Debug) {
        printf("GpibGS::write cmdGpib error after count %d\n",count);
        return(-1);
    }
    return(actual);
}

gpibStatus GpibGS::read(char *buf, int cnt, int *actual,
    int addr, double time_limit, int eos)
{
    *actual=0; *buf=0;
    char cmdbuf[3];
    strcpy(cmdbuf,untalkUnlisten);
    cmdbuf[2] = (char)(addr + 0x40);
    int count = cmdGpib(cmdbuf,3,time_limit);
    if(count != 3 && GpibGsTi9914Debug) {
        printf("GpibGS::read cmdGpib error before count %d\n",count);
        return(gpibTimeout);
    }
    *actual = readGpib(buf,cnt,time_limit,eos);
    count = cmdGpib(untalkUnlisten, 2, time_limit);
    if(count != 2 && GpibGsTi9914Debug) {
        printf("GpibGS::read cmdGpib error after count %d actual %d\n",
           count,*actual);
        return(gpibTimeout);
    }
    if(GpibGsTi9914Debug){
        printf("GpibGS::read addr %d cnt %d time_limit %f eos 0x%2.2x[",
            addr,cnt,time_limit,eos);
        for (int i=0; i< *actual; i++) printf("%3.2x",buf[i]);
        printf("]\n");
    }
    return(gpibSuccess);
}

void GpibGS::interruptHandler(void* v)
{
    GpibGS *pGpib = (GpibGS *)v;
    int message_complete = 0;
    unsigned char newchar = 0;
    pGpib->copyIsr0 |= pGpib->regs->intReg0;
    pGpib->copyIsr1 |= pGpib->regs->intReg1;
    switch (pGpib->transferState)
    {
    case IP488_READ:
        if (pGpib->copyIsr0 & isr0BI)
        { // We are ready to read a byte now
            if (pGpib->transferSize == 0)
            { // We are out of buffer space; throw character away
                newchar = pGpib->regs->data;
                pGpib->transferState = IP488_IDLE;
                pGpib->ioStatus = gpibFailure;	// buffer overflow
		semGive(pGpib->ioComplete);
                return ;
            }
            *pGpib->transferBuffer = pGpib->regs->data;
            if (pGpib->transferSize == 1)
            { // This is the last character which will fit in buffer
                message_complete = 1;
            }
            else if (pGpib->transferEos != -1 ) 
            { // Check for EOS character
                if (*pGpib->transferBuffer == pGpib->transferEos) {
                    // This is EOS character, message complete
                    message_complete = 1;
                }
            }
            if(pGpib->ph) (*pGpib->ph)(pGpib->phParm,true,*pGpib->transferBuffer);
            --(pGpib->transferSize);
            ++(pGpib->transferBuffer);
            pGpib->copyIsr0 &= ~isr0BI;
        }
        if (pGpib->copyIsr0 & isr0END)
        { // EOI was asserted for this character
            pGpib->copyIsr0 &= ~isr0END;
            message_complete = 1;
        }
        if (message_complete)
        {
            cmdIP488_AUXlonC(pGpib->regs);
            pGpib->transferState = IP488_IDLE;
	    pGpib->ioStatus = gpibSuccess;
	    semGive(pGpib->ioComplete);
            return ;
        }
        break;

    case IP488_WRITE:
        // If data transmitter is ready, send next byte
        if (pGpib->copyIsr1 & isr1ERR)
        { // Handshake failure -- we ain't got a listener
            pGpib->copyIsr1 &= ~isr1ERR;		// Clear error status
            pGpib->regs->cmdStatus = IP488_AUXnbaf; //throw away last byte
            ++(pGpib->transferSize);			// Clean up
            --(pGpib->transferBuffer);			// Clean up
            logMsg("GpibGsTi9914::interruptHandler Handshake failure\n",
                0,0,0,0,0,0);
            pGpib->transferState = IP488_IDLE;
	    pGpib->ioStatus = gpibFailure;
	    semGive(pGpib->ioComplete);
            return ;
        }
        if (pGpib->copyIsr0 & isr0BO)
        { // We are ready to send a byte now
            if (pGpib->transferSize == 0) {
                cmdIP488_AUXtonC(pGpib->regs);
                pGpib->transferState = IP488_IDLE;
	        pGpib->ioStatus = gpibSuccess;
	        semGive(pGpib->ioComplete);
                return ;
            }
            if(pGpib->ph)
                (*pGpib->ph)(pGpib->phParm,false,*pGpib->transferBuffer);
            if ((pGpib->transferSize == 1) && pGpib->transferEoiOnLast) {
                pGpib->regs->cmdStatus = IP488_AUXfeoi;
            }
            pGpib->regs->data =
                (unsigned short)(unsigned char)(*pGpib->transferBuffer);
            --(pGpib->transferSize);
            ++(pGpib->transferBuffer);

            pGpib->copyIsr0 &= ~isr0BO;
        }
        break;
    case IP488_IDLE:
        logMsg("GpibGsTi9914::interrupt IP488_IDLE isr0BI %d isr0BO %d\n",
            ((pGpib->copyIsr0 & isr0BI) ? 1 : 0),
            ((pGpib->copyIsr0 & isr0BO) ? 1 : 0),0,0,0,0);
        break;
    }
    return ;
}

void GpibGS::sicGpib(void)
{
    if(GpibGsTi9914Debug) printf("GpibGS::sicGpib\n");
    transferState = IP488_IDLE;
    regs->cmdStatus = IP488_AUXsicS;
    taskDelay(1);
    regs->cmdStatus = IP488_AUXsicC;
}

// Write a string with the ATN line active
int GpibGS::cmdGpib(const char *buf, int cnt, double time_limit)
{
    if (cnt == 0) {
        printf("GpibGS::cmdGpib called with cnt=0\n");
        ioStatus = gpibSuccess;
        return (0);
    }
    // Send the first byte manually
    int lockKey = ::intLock();
    if(!cmdIP488_AUXtca(regs)) {
        ::intUnlock(lockKey);
        return(0);
    }
    ioStatus = gpibSuccess;
    transferEoiOnLast = false;
    transferSize = cnt;
    transferBuffer = (unsigned char *)buf;
    transferState = IP488_WRITE;
    interruptHandler(this);
    ::intUnlock(lockKey);
    waitIo(time_limit);
    if(ioStatus==gpibTimeout) sicGpib();
    return(cnt - transferSize);
}

int GpibGS::readGpib(char *buf, int cnt, double time_limit, int eos)
{
    if (cnt == 0) {
        printf("GpibGS::readGpib called with cnt=0\n");
        ioStatus = gpibSuccess;
        return (0);
    }
    int lockKey = ::intLock();
    regs->cmdStatus = IP488_AUXrhdf;
    regs->cmdStatus = IP488_AUXhdfeS; // Holdoff on EOI
    if(!cmdIP488_AUXlonS(regs) || !cmdIP488_AUXgts(regs)) {
        ::intUnlock(lockKey);
        return(0);
    }
    ioStatus = gpibSuccess;
    transferSize = cnt;
    transferBuffer = (unsigned char *)buf;
    transferState = IP488_READ;
    transferEos = eos;
    ::intUnlock(lockKey);
    waitIo(time_limit);
    if(ioStatus==gpibTimeout) sicGpib();
    cmdIP488_AUXlonC(regs);
    return(cnt - transferSize);
}

int GpibGS::writeGpib(const char *buf, int cnt, bool raw_flag, double time_limit)
{
    if (cnt == 0) {
        printf("GpibGS::writeGpib called with cnt=0\n");
        ioStatus = gpibSuccess;
        return (0);
    }
    int lockKey = ::intLock();
    ioStatus = gpibSuccess;
    if(!cmdIP488_AUXgts(regs) || !cmdIP488_AUXtonS(regs)) {
        ::intUnlock(lockKey);
        return(0);
    }
    if(raw_flag) 
	transferEoiOnLast = false;
    else
	transferEoiOnLast = true;
    transferSize = cnt;
    transferBuffer = (unsigned char *)buf;
    transferState = IP488_WRITE;
    ::intUnlock(lockKey);
    waitIo(time_limit);
    if(ioStatus==gpibTimeout) sicGpib();
    cmdIP488_AUXtonC(regs);
    return(cnt - transferSize);

}


void GpibGS::rebootCallback(void *v)
{
    GpibGS *pGpibGS = (GpibGS *)v;
    pGpibGS->pIPM->intDisable(0);
}
