//SerialPort.cc

/*************************************************************************
* Copyright (c) 2002 The University of Chicago, as Operator of Argonne
* National Laboratory, and the Regents of the University of California, as
* Operator of Los Alamos National Laboratory. EPICS BASE Versions 3.13.7
* and higher are distributed subject to a Software License Agreement found
* in file LICENSE that is included with this distribution.
*************************************************************************/
/*
	Current Author: Marty Kraimer, Thomas Coleman
	Original Author: Jim Kowalkowski (for Hideos)
	Date: 09JUL98
*/
#include <stdlib.h>
#include <stddef.h>
#include <string.h>
#include <stdio.h>
 
#include <semLib.h>
#include <tickLib.h>
#include <taskLib.h>
#include <intLib.h>
 
#include "WatchDog.h"
#include "DLList.h"
#include "SerialPort.h"
 
static DLList *pSerialPortList = 0;

SerialPort:: SerialPort()
: request(requestNone), ioStatus(serialPortSuccess),
  bh(0), bhParm(0), ph(0), phParm(0), ioComplete(0), pwg(0),
  isBound(false), portName(0), clientStartedIoTimeout(false)
{}

int SerialPort::registerName(const char *name)
{
    if(portName!=0) {
        printf("SerialPort::registerName already registered %s\n",name);
        return(-1);
    }
    ioComplete = semBCreate(SEM_Q_FIFO, SEM_EMPTY);
    pwg = new WatchDog;
    portName = name;
    if(pSerialPortList==0) pSerialPortList = new DLList;
    DLNode *node = DLList::allocNode();
    node->obj = this;
    pSerialPortList->addEnd(node);
    return(0);
}

SerialPort *SerialPort::find(const char *portName)
{
    if(pSerialPortList==0) pSerialPortList = new DLList;
    DLNode *node = pSerialPortList->getFirst();
    while(node) {
        SerialPort *pSerialPort = (SerialPort *)node->obj;
        if(::strcmp(portName,pSerialPort->portName)==0) {
            return(pSerialPort);
        }
        node = pSerialPortList->getNext(node);
    }
    return(0);
}

SerialPort *SerialPort::bind(const char *portName,void* parm,BYTE_HANDLER b)
{
    SerialPort *pSerialPort = find(portName);
    if(!pSerialPort) {
	printf("%s port not found\n",portName);
	return(0);
    }
    if(pSerialPort->isBound) {
        printf("%s port already bound\n",portName);
        return(0);
    }
    if(b==0) {
	printf("%s bind call but no byte handler supplied",portName);
	return(0);
    }
    pSerialPort->bhParm=parm;
    pSerialPort->bh=b;
    pSerialPort->isBound = true;
    return(pSerialPort);
}

void SerialPort::release()
{
    bh=0; bhParm=0; isBound = false;
}

bool SerialPort::setSniffHandler(
    const char *portName,void* phParm,PEEK_HANDLER ph)
{
    SerialPort *pSerialPort = find(portName);
    if(!pSerialPort) {
	printf("%s port not found\n",portName);
	return(false);
    }
    if(!pSerialPort->isBound) {
        printf("%s port not bound\n",portName);
        return(false);
    }
    if(pSerialPort->ph!=0) {
	printf("%s already has a PEEK_HANDLER\n",portName);
	return(0);
    }
    int lockKey = intLock();
    pSerialPort->ph = ph;
    pSerialPort->phParm = phParm;
    intUnlock(lockKey);
    return(true);
}

void SerialPort::freeSniffHandler(const char *portName)
{
    SerialPort *pSerialPort = find(portName);
    if(!pSerialPort) {
	printf("%s port not found\n",portName);
	return;
    }
    int lockKey = intLock();
    pSerialPort->ph = 0;
    pSerialPort->phParm = 0;
    intUnlock(lockKey);
    return;
}

void SerialPort::clientStartIoTimeout(int seconds)
{
    startIoTimeout(seconds);
    clientStartedIoTimeout = true;
}

void SerialPort::clientStartIoTimeout(double seconds)
{
    startIoTimeout(seconds);
    clientStartedIoTimeout = true;
}

void SerialPort::startIoTimeout(int seconds)
{
    if(clientStartedIoTimeout) return;
    // Make sure there is no lingering semGive
    semTake(ioComplete,NO_WAIT);
    pwg->start(seconds,timoutHandler,this);
}

void SerialPort::startIoTimeout(double seconds)
{
    if(clientStartedIoTimeout) return;
    // Make sure there is no lingering semGive
    semTake(ioComplete,NO_WAIT);
    pwg->start(seconds,timoutHandler,this);
}

void SerialPort::waitIo()
{
    semTake(ioComplete,WAIT_FOREVER);
    request = requestNone;
    clientStartedIoTimeout = false;
    pwg->cancel();
}

void SerialPort::timoutHandler(void *pvt)
{
    SerialPort *pSerialPort = (SerialPort *)pvt;
    pSerialPort->ioStatus = serialPortTimeout;
    semGive(pSerialPort->ioComplete);
}
